from typing import Dict, Any, List

from mockfirestore._helpers import get_document_iterator, get_by_path, set_by_path, delete_by_path


def apply_transformations(document: Dict[str, Any], data: Dict[str, Any]):
    """Handles special fields like INCREMENT."""
    increments = {}
    arr_unions = {}
    arr_deletes = {}
    deletes = []

    for key, value in list(get_document_iterator(data)):
        if not value.__class__.__module__.startswith('google.cloud.firestore'):
            # Unfortunately, we can't use `isinstance` here because that would require
            # us to declare google-cloud-firestore as a dependency for this library.
            # However, it's somewhat strange that the mocked version of the library
            # requires the library itself, so we'll just leverage this heuristic as a
            # means of identifying it.
            #
            # Furthermore, we don't hardcode the full module name, since the original
            # library seems to use a thin shim to perform versioning. e.g. at the time
            # of writing, the full module name is `google.cloud.firestore_v1.transforms`,
            # and it can evolve to `firestore_v2` in the future.
            continue

        transformer = value.__class__.__name__
        if transformer == 'Increment':
            increments[key] = value.value
        elif transformer == 'ArrayUnion':
            arr_unions[key] = value.values
        elif transformer == 'ArrayRemove':
            arr_deletes[key] = value.values
            del data[key]
        elif transformer == 'Sentinel':
            if value.description == "Value used to delete a field in a document.":
                deletes.append(key)
                del data[key]

        # All other transformations can be applied as needed.
        # See #29 for tracking.

    def _update_data(new_values: dict, default: Any):
        for key, value in new_values.items():
            path = key.split('.')

            try:
                item = get_by_path(document, path)
            except (TypeError, KeyError):
                item = default

            set_by_path(data, path, item + value, create_nested=True)

    _update_data(increments, 0)
    _update_data(arr_unions, [])

    _apply_updates(document, data)
    _apply_deletes(document, deletes)
    _apply_arr_deletes(document, arr_deletes)


def _apply_updates(document: Dict[str, Any], data: Dict[str, Any]):
    for key, value in data.items():
        path = key.split(".")
        set_by_path(document, path, value, create_nested=True)


def _apply_deletes(document: Dict[str, Any], data: List[str]):
    for key in data:
        path = key.split(".")
        delete_by_path(document, path)


def _apply_arr_deletes(document: Dict[str, Any], data: Dict[str, Any]):
    for key, values_to_delete in data.items():
        path = key.split(".")
        try:
            value = get_by_path(document, path)
        except KeyError:
            continue
        for value_to_delete in values_to_delete:
            try:
                value.remove(value_to_delete)
            except ValueError:
                pass
        set_by_path(document, path, value)


from unittest import TestCase
from mockfirestore import MockFirestore, Transaction


class TestTransaction(TestCase):
    def setUp(self) -> None:
        self.fs = MockFirestore()
        self.fs._data = {'foo': {
                'first': {'id': 1},
                'second': {'id': 2}
            }}

    def test_transaction_getAll(self):
        with Transaction(self.fs) as transaction:
            transaction._begin()
            docs = [self.fs.collection('foo').document(doc_name)
                    for doc_name in self.fs._data['foo']]
            results = list(transaction.get_all(docs))
            returned_docs_snapshots = [result.to_dict() for result in results]
            expected_doc_snapshots = [doc.get().to_dict() for doc in docs]
            for expected_snapshot in expected_doc_snapshots:
                self.assertIn(expected_snapshot, returned_docs_snapshots)

    def test_transaction_getDocument(self):
        with Transaction(self.fs) as transaction:
            transaction._begin()
            doc = self.fs.collection('foo').document('first')
            returned_doc = next(transaction.get(doc))
            self.assertEqual(doc.get().to_dict(), returned_doc.to_dict())

    def test_transaction_getQuery(self):
        with Transaction(self.fs) as transaction:
            transaction._begin()
            query = self.fs.collection('foo').order_by('id')
            returned_docs = [doc.to_dict() for doc in transaction.get(query)]
            query = self.fs.collection('foo').order_by('id')
            expected_docs = [doc.to_dict() for doc in query.stream()]
            self.assertEqual(returned_docs, expected_docs)

    def test_transaction_set_setsContentOfDocument(self):
        doc_content = {'id': '3'}
        doc_ref = self.fs.collection('foo').document('third')
        with Transaction(self.fs) as transaction:
            transaction._begin()
            transaction.set(doc_ref, doc_content)
        self.assertEqual(doc_ref.get().to_dict(), doc_content)

    def test_transaction_set_mergeNewValue(self):
        doc = self.fs.collection('foo').document('first')
        with Transaction(self.fs) as transaction:
            transaction._begin()
            transaction.set(doc, {'updated': True}, merge=True)
        updated_doc = {'id': 1, 'updated': True}
        self.assertEqual(doc.get().to_dict(), updated_doc)

    def test_transaction_update_changeExistingValue(self):
        doc = self.fs.collection('foo').document('first')
        with Transaction(self.fs) as transaction:
            transaction._begin()
            transaction.update(doc, {'updated': False})
        updated_doc = {'id': 1, 'updated': False}
        self.assertEqual(doc.get().to_dict(), updated_doc)

    def test_transaction_delete_documentDoesNotExistAfterDelete(self):
        doc = self.fs.collection('foo').document('first')
        with Transaction(self.fs) as transaction:
            transaction._begin()
            transaction.delete(doc)
        doc = self.fs.collection('foo').document('first').get()
        self.assertEqual(False, doc.exists)






from functools import partial
import random
from typing import Iterable, Callable
from mockfirestore._helpers import generate_random_string, Timestamp
from mockfirestore.document import DocumentReference, DocumentSnapshot
from mockfirestore.query import Query

MAX_ATTEMPTS = 5
_MISSING_ID_TEMPLATE = "The transaction has no transaction ID, so it cannot be {}."
_CANT_BEGIN = "The transaction has already begun. Current transaction ID: {!r}."
_CANT_ROLLBACK = _MISSING_ID_TEMPLATE.format("rolled back")
_CANT_COMMIT = _MISSING_ID_TEMPLATE.format("committed")


class WriteResult:
    def __init__(self):
        self.update_time = Timestamp.from_now()


class Transaction:
    """
    This mostly follows the model from
    https://googleapis.dev/python/firestore/latest/transaction.html
    """
    def __init__(self, client,
                 max_attempts=MAX_ATTEMPTS, read_only=False):
        self._client = client
        self._max_attempts = max_attempts
        self._read_only = read_only
        self._id = None
        self._write_ops = []
        self.write_results = None

    @property
    def in_progress(self):
        return self._id is not None

    @property
    def id(self):
        return self._id

    def _begin(self, retry_id=None):
        # generate a random ID to set the transaction as in_progress
        self._id = generate_random_string()

    def _clean_up(self):
        self._write_ops.clear()
        self._id = None

    def _rollback(self):
        if not self.in_progress:
            raise ValueError(_CANT_ROLLBACK)

        self._clean_up()

    def _commit(self) -> Iterable[WriteResult]:
        if not self.in_progress:
            raise ValueError(_CANT_COMMIT)

        results = []
        for write_op in self._write_ops:
            write_op()
            results.append(WriteResult())
        self.write_results = results
        self._clean_up()
        return results

    def get_all(self,
                references: Iterable[DocumentReference]) -> Iterable[DocumentSnapshot]:
        return self._client.get_all(references)

    def get(self, ref_or_query) -> Iterable[DocumentSnapshot]:
        if isinstance(ref_or_query, DocumentReference):
            return self._client.get_all([ref_or_query])
        elif isinstance(ref_or_query, Query):
            return ref_or_query.stream()
        else:
            raise ValueError(
                'Value for argument "ref_or_query" must be a DocumentReference or a Query.'
            )

    # methods from
    # https://googleapis.dev/python/firestore/latest/batch.html#google.cloud.firestore_v1.batch.WriteBatch

    def _add_write_op(self, write_op: Callable):
        if self._read_only:
            raise ValueError(
                "Cannot perform write operation in read-only transaction."
            )
        self._write_ops.append(write_op)

    def create(self, reference: DocumentReference, document_data):
        # this is a no-op, because if we have a DocumentReference
        # it's already in the MockFirestore
        ...

    def set(self, reference: DocumentReference, document_data: dict,
            merge=False):
        write_op = partial(reference.set, document_data, merge=merge)
        self._add_write_op(write_op)

    def update(self, reference: DocumentReference,
               field_updates: dict, option=None):
        write_op = partial(reference.update, field_updates)
        self._add_write_op(write_op)

    def delete(self, reference: DocumentReference, option=None):
        write_op = reference.delete
        self._add_write_op(write_op)

    def commit(self):
        return self._commit()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is None:
            self.commit()



from unittest import TestCase

from mockfirestore import MockFirestore


class TestDocumentSnapshot(TestCase):
    def test_documentSnapshot_toDict(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'id': 1}
        }}
        doc = fs.collection('foo').document('first').get()
        self.assertEqual({'id': 1}, doc.to_dict())

    def test_documentSnapshot_toDict_isolation(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'id': 1}
        }}
        doc_dict = fs.collection('foo').document('first').get().to_dict()
        fs._data['foo']['first']['id'] = 2
        self.assertEqual({'id': 1}, doc_dict)

    def test_documentSnapshot_exists(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'id': 1}
        }}
        doc = fs.collection('foo').document('first').get()
        self.assertTrue(doc.exists)

    def test_documentSnapshot_exists_documentDoesNotExist(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'id': 1}
        }}
        doc = fs.collection('foo').document('second').get()
        self.assertFalse(doc.exists)

    def test_documentSnapshot_reference(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'id': 1}
        }}
        doc_ref = fs.collection('foo').document('second')
        doc_snapshot = doc_ref.get()
        self.assertIs(doc_ref, doc_snapshot.reference)

    def test_documentSnapshot_id(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'id': 1}
        }}
        doc = fs.collection('foo').document('first').get()
        self.assertIsInstance(doc.id, str)

    def test_documentSnapshot_create_time(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'id': 1}
        }}
        doc = fs.collection('foo').document('first').get()
        self.assertIsNotNone(doc.create_time)

    def test_documentSnapshot_update_time(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'id': 1}
        }}
        doc = fs.collection('foo').document('first').get()
        self.assertIsNotNone(doc.update_time)

    def test_documentSnapshot_read_time(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'id': 1}
        }}
        doc = fs.collection('foo').document('first').get()
        self.assertIsNotNone(doc.read_time)

    def test_documentSnapshot_get_by_existing_field_path(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'id': 1, 'contact': {
                'email': 'email@test.com'
            }}
        }}
        doc = fs.collection('foo').document('first').get()
        self.assertEqual(doc.get('contact.email'), 'email@test.com')

    def test_documentSnapshot_get_by_non_existing_field_path(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'id': 1, 'contact': {
                'email': 'email@test.com'
            }}
        }}
        doc = fs.collection('foo').document('first').get()
        with self.assertRaises(KeyError):
            doc.get('contact.phone')

    def test_documentSnapshot_get_in_an_non_existing_document(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'id': 1, 'contact': {
                'email': 'email@test.com'
            }}
        }}
        doc = fs.collection('foo').document('second').get()
        self.assertIsNone(doc.get('contact.email'))

    def test_documentSnapshot_get_returns_a_copy_of_the_data_stored(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'id': 1, 'contact': {
                'email': 'email@test.com'
            }}
        }}
        doc = fs.collection('foo').document('first').get()
        self.assertIsNot(
            doc.get('contact'),fs._data['foo']['first']['contact']
        )
